/*************************************************************************
* Simple TFTP client by J. Fustin
*
* This program was written as an assignment for Computer Networks course
* at FER (www.fer.hr/english). Its purpose is to be as simple as possible
* so everyone could learn how to perform basic communication with TFTP
* server over UDP/IP.
*
* Change log:
* 01/2005 - initial release
* 03/2007 - refactored a bit
* 11/2009 - cleaned up a bit
*************************************************************************/

#include "common.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <inttypes.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>

void print_instructions()
{
	const char* instructions =
		"Simple TFTP client by J. Fustin.\n"
		"Usage: client <get|put> <host> <port> <file>\n";
	
	fprintf(stderr, "%s", instructions);
}

void fatal_error(const char* message)
{
	fprintf(stderr, "Fatal error: %s\n", message);
	exit(1);
}

void sigint_handler(int sig)
{
	printf("Bye bye.\n");
	exit(1);
}

/*************************************************************************
* [in] filename, sck, addr
*************************************************************************/
void get_file(const char* filename, int sck, struct sockaddr_in* addr)
{
	int rf;		/* number of bytes read from socket */
	int sel;	/* return value from select() */
	
	FILE* file;
	
	unsigned int addrlen = sizeof(*addr);
	
	unsigned int retrynum;
	
	char buffer[BUFF_SIZE];
	uint16_t blocknum;
	uint16_t rcv_opcode;
	uint16_t rcv_blocknum;
	
	/* See if we can create the file. */
	file = fopen(filename, "w");
	if(!file)
		fatal_error("Can not create a file with that name.");
	
	/* Initial state. */
	retrynum = 0;
	blocknum = 0;
	
	while(1)
	{
		if(retrynum > 5)
		{
			printf("Too many retries. Aborting transfer!\n");
			break;
		}
		
		/* 1. a) Send RRQ frame. */
		if(blocknum == 0)
		{
			make_rrq_frame(filename, buffer);
			sendto(sck, buffer, 2 + strlen(filename) + 1 + strlen("octet") + 1,
				0, (struct sockaddr *) addr, addrlen);
		}
		/* 1. b) Send ACK frame. */
		else
		{
			make_ack_frame(blocknum, buffer);
			sendto(sck, buffer, 4, 0, (struct sockaddr *) addr, addrlen);
		}
		
		/* 2. Wait for new DATA frame. */
		sel = poll_socket(sck, 2);
		
		if(sel == -1 || sel == 0)
		{
			++retrynum;
			continue;
		}
		
		rf = recvfrom(sck, buffer, BUFF_SIZE, 0, (struct sockaddr *) addr, &addrlen);
		parse_frame(buffer, &rcv_opcode, &rcv_blocknum);
		
		if(rcv_opcode == OPCODE_ERROR)
		{
			printf("Server error: %s. Aborting transfer!\n", buffer + 4);
			break;
		}
		else if(rcv_opcode != OPCODE_DATA || rcv_blocknum != blocknum + 1)
		{
			++retrynum;
			continue;
		}
		
		/* Everything is OK! */
		++blocknum;
		retrynum = 0;
		
		fwrite(buffer + 4, 1, rf - 4, file);
		
		/* End of transfer? */
		if(rf < 516)
		{
			make_ack_frame(blocknum, buffer);
			sendto(sck, buffer, 4, 0, (struct sockaddr *) addr, addrlen);
			printf("Transfer completed (received %d bytes).\n", (blocknum-1) * 512 + rf - 4);
			break;
		}
	}
	
	fclose(file);
}

/*************************************************************************
* [in] filename, sck, addr
*************************************************************************/
void put_file(const char* filename, int sck, struct sockaddr_in* addr)
{
	int rd = 0;	/* number of bytes read from file */
	int rf;		/* number of bytes read from socket */
	int sel;	/* return value from select() */
	
	FILE* file;
	
	unsigned int addrlen = sizeof(*addr);
	
	unsigned int readnext;
	unsigned int retrynum;
	
	char buffer[BUFF_SIZE];
	uint16_t blocknum;
	
	char rcv_buffer[BUFF_SIZE];
	uint16_t rcv_opcode;
	uint16_t rcv_blocknum;
	
	/* See if the file exists. */
	file = fopen(filename, "r");
	if(!file)
		fatal_error("File not found.");
	
	/* Initial state. */
	blocknum = 0;
	retrynum = 0;
	readnext = 1;
	
	while(1)
	{
		if(retrynum > 5)
		{
			printf("Too many retries. Aborting transfer!\n");
			break;
		}
		
		/* 1. a) Send WRQ frame. */
		if(blocknum == 0)
		{
			make_wrq_frame(filename, buffer);
			sendto(sck, buffer, 2 + strlen(filename) + 1 + strlen("octet") + 1,
				0, (struct sockaddr *) addr, addrlen);
		}
		/* 1. b) Send DATA frame. */
		else
		{
			if(readnext)
			{
				make_data_frame(blocknum, buffer);
				rd = fread(buffer + 4, 1, 512, file);
				readnext = 0;
			}
			
			sendto(sck, buffer, 4 + rd, 0, (struct sockaddr *) addr, addrlen);
		}
		    
		/* 2. Wait for ACK frame. */
		sel = poll_socket(sck, 2);
		
		if(sel == -1 || sel == 0)
		{
			++retrynum;
			continue;
		}
		
		rf = recvfrom(sck, rcv_buffer, BUFF_SIZE, 0, (struct sockaddr *) addr, &addrlen);
		parse_frame(rcv_buffer, &rcv_opcode, &rcv_blocknum);
		
		if(rcv_opcode == OPCODE_ERROR)
		{
			printf("Server error: %s. Aborting transfer!\n", rcv_buffer + 4);
			break;
		}
		else if(rcv_opcode != OPCODE_ACK || rcv_blocknum != blocknum)
		{
			++retrynum;
			continue;
		}
		    
		/* Everything is OK! */
		++blocknum;
		retrynum = 0;
		readnext = 1;
		
		/* End of transfer? */
		if(rd < 512 && blocknum > 1)
		{
			printf("Transfer completed (sent %d bytes).\n", (blocknum-2) * 512 + rd);
			break;
		}
	}
	
	fclose(file);
}

/*************************************************************************
* [in] action, hostip, port, filename
*************************************************************************/
void start_tftp_client(action act, const struct in_addr *hostip,
	unsigned short port, const char *filename)
{
	int sck;					/* socket descriptor */
	struct sockaddr_in addr;	/* host's address struct */
	
	/* Initialize host's address struct. */
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	memcpy(&addr.sin_addr, hostip, sizeof(*hostip));
	
	/* Create a connectionless (UDP/IP) socket. */
	sck = socket(AF_INET, SOCK_DGRAM, 0);
	
	if(sck == -1)
		fatal_error("Can not create socket.");
	
	switch(act)
	{
		case ACTION_RECEIVE:
			get_file(filename, sck, &addr);
			break;
			
		case ACTION_SEND:
			put_file(filename, sck, &addr);
			break;
			
		case ACTION_UNKNOWN:
			/* This shouldn't happen. */
			break;
	}
	
	close(sck);
}

int main(int argc, char *argv[])
{
	action act;
	
	unsigned short port;
	
	struct hostent* ent;
	struct in_addr hostip;
	
	if(argc != 5)
	{
		print_instructions();
		exit(1);
	}
	
	/* Parse argv[1]. */
	if(!strcmp("get", argv[1]))
	{
		act = ACTION_RECEIVE;
	}
	else if(!strcmp("put", argv[1]))
	{
		act = ACTION_SEND;
	}
	else
	{
		print_instructions();
		exit(1);
	}
	
	/* Parse argv[2]. */
	ent = gethostbyname(argv[2]);
	if(!ent)
	{
		fatal_error("Bad host name.");
	}
	else
	{
		memcpy(&hostip, ent->h_addr_list[0], ent->h_length);
	}
	
	/* Parse argv[3]. */
	port = atoi(argv[3]);
	
	signal(SIGINT, sigint_handler);
	
	start_tftp_client(act, &hostip, port, argv[4]);
	
	return 0;
}

